/* BEGIN software license
 *
 * msXpertSuite - mass spectrometry software suite
 * -----------------------------------------------
 * Copyright(C) 2009,...,2021 Filippo Rusconi
 *
 * http://www.msxpertsuite.org
 *
 * This file is part of the msXpertSuite project.
 *
 * The msXpertSuite project is the successor of the massXpert project. This
 * project now includes various independent modules:
 *
 * - massXpert, model polymer chemistries and simulate mass spectrometric data;
 * - mineXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * END software license
 */


#include <qmath.h>

#include <limits>
#include <iterator>

#include <QVector>
#include <QDebug>

#include "../../exception/exceptionnotrecognized.h"

#include "filterlowintensitysignalremoval.h"


namespace pappso
{


FilterLowIntensitySignalRemoval::FilterLowIntensitySignalRemoval(
  double mean, double std_dev, double threshold)
{
  m_noiseMean   = mean;
  m_noiseStdDev = std_dev;
  m_threshold   = threshold;
}


FilterLowIntensitySignalRemoval::FilterLowIntensitySignalRemoval(
  const QString &parameters)
{
  buildFilterFromString(parameters);
}


FilterLowIntensitySignalRemoval::FilterLowIntensitySignalRemoval(
  const FilterLowIntensitySignalRemoval &other)
{
  m_noiseMean   = other.m_noiseMean;
  m_noiseStdDev = other.m_noiseStdDev;
  m_threshold   = other.m_threshold;
}


FilterLowIntensitySignalRemoval::~FilterLowIntensitySignalRemoval()
{
}


FilterLowIntensitySignalRemoval &
FilterLowIntensitySignalRemoval::operator=(
  const FilterLowIntensitySignalRemoval &other)
{
  if(&other == this)
    return *this;

  m_noiseMean   = other.m_noiseMean;
  m_noiseStdDev = other.m_noiseStdDev;
  m_threshold   = other.m_threshold;

  return *this;
}


void
FilterLowIntensitySignalRemoval::buildFilterFromString(
  const QString &parameters)
{
  // Typical string: "FloorAmplitudePercentage|15"
  if(parameters.startsWith(QString("%1|").arg(name())))
    {
      QStringList params = parameters.split("|").back().split(";");

      m_noiseMean   = params.at(0).toDouble();
      m_noiseStdDev = params.at(1).toDouble();
      m_threshold   = params.at(2).toDouble();
    }
  else
    {
      throw pappso::ExceptionNotRecognized(
        QString(
          "Building of FilterLowIntensitySignalRemoval from string %1 failed")
          .arg(parameters));
    }
}


std::size_t
FilterLowIntensitySignalRemoval::detectIsotopicClusters(const Trace &trace)
{
  // We want to iterate in the trace and check if we can get the clusters
  // isolated one by one.

  m_isotopicClusters.clear();

  std::size_t trace_size = trace.size();
  if(trace_size <= 2)
    {
      qDebug() << "The original trace has less than 3 points. Returning it "
                  "without modification.";
      return m_isotopicClusters.size();
    }
  else
    qDebug() << "The original trace has" << trace_size << "data points";

  std::size_t index = 0;

  // Seed the system with the first point of the trace.
  m_prev = IndexIntensity(index, trace[index].y);

  qDebug() << "First trace point: "
           << "(" << trace[m_prev.first].x << "," << m_prev.second << ");";

  // We still have not found any apex!
  m_prevApex = std::pair(0, nan);

  // We will need to know if we were ascending to an apex.
  bool was_ascending_to_apex = false;

  ClusterSPtr cluster_sp = std::make_shared<Cluster>();

  // Now that we have seeded the system with the first point of the original
  // trace, go to the next one.
  ++index;

  // And now iterate in the original trace and make sure we detect all the
  // clusters.

  while(index < trace_size)
    {
      // Get the current trace point as an IndexIntensity.
      m_cur = IndexIntensity(index, trace[index].y);

      qDebug() << "Current trace point: "
               << "(" << trace[m_cur.first].x << "," << m_cur.second << ");";

      // We monitor if we are going up or down a peak.

      if(m_cur.second > m_prev.second)
        // We are ascending a peak. We do not know if we are at the apex.
        {
          qDebug().noquote() << "We are ascending to an apex.\n";
          was_ascending_to_apex = true;
        }
      else
        // We are descending a peak.
        {
          qDebug().noquote() << "Descending a peak. ";

          // There are two situations:
          //
          // 1. Either we were ascending to an apex, and m_prev is that apex,
          //
          // 2. Or we were not ascending to an apex and in fact all we are doing
          // is going down an apex that occurred more than one trace point ago.

          if(!was_ascending_to_apex)
            // We were not ascending to an apex.
            {
              // Do nothing here.

              qDebug().noquote()
                << "But, we were not ascending to an apex, so do nothing.\n";
            }
          else
            // We are effectively descending a peak right after the apex was
            // reached.
            {
              qDebug().noquote()
                << "And, we were ascending to an apex, so "
                   "m_curApeex becomes m_prev: ("
                << trace[m_curApex.first].x << "," << m_curApex.second << ")\n";

              m_curApex = m_prev;

              // We might have two situations:

              // 1. We had already encountered an apex.
              // 2. We had not yet encountered an apex.

              if(!std::isnan(m_prevApex.second))
                // We had already encountered an apex.
                {
                  // Was that apex far on the left of the current apex ?

                  if(trace[m_curApex.first].x - trace[m_prevApex.first].x >
                     INTRA_CLUSTER_INTER_PEAK_DISTANCE)
                    // The distance is not compatible with both apices to belong
                    // to the same cluster.
                    {
                      // We are creating a new isotopic cluster.

                      // But, since another apex had been encountered already,
                      // that means that an isotopic cluster was cooking
                      // already. We must store it.

                      if(!cluster_sp->size())
                        qFatal("Cannot be that the cluster has no apex.");

                      m_isotopicClusters.push_back(cluster_sp);

                      qDebug().noquote()
                        << "There was a previous apex already, BUT "
                           "outside of cluster range. "
                           "Pushing the cooking cluster that has size:"
                        << cluster_sp->size();

                      // Now create a brand new cluster.
                      cluster_sp = std::make_shared<Cluster>();

                      qDebug() << "Created a brand new cluster.";

                      // We only start the new cluster with the current apex if
                      // that apex is above the threshold.

                      if(m_curApex.second > m_threshold)
                        {
                          // And start it with the current apex.
                          cluster_sp->push_back(
                            std::make_shared<IndexIntensity>(m_curApex));

                          qDebug()
                            << "Since the current apex is above the threshold, "
                               "we PUSH it to the newly created cluster: ("
                            << trace[m_curApex.first].x << ","
                            << m_curApex.second << ")";

                          m_prevApex = m_curApex;

                          qDebug() << "Set prev apex to be cur apex.";
                        }
                      else
                        {
                          qDebug()
                            << "Since the current apex is below the threshold, "
                               "we DO NOT push it to the newly created "
                               "cluster: ("
                            << trace[m_curApex.first].x << ","
                            << m_curApex.second << ")";

                          // Since previous apex went to the closed cluster, we
                          // need to reset it.

                          m_prevApex = std::pair(0, nan);

                          qDebug() << "Since the previous apex went to the "
                                      "closed cluster, and cur apex has too "
                                      "small an intensity, we reset prev apex "
                                      "to std::pair(0, nan).";
                        }
                    }
                  else
                    // The distance is compatible with both apices to belong to
                    // the same isotopic cluster.
                    {

                      // But we only push back the current apex to the cluster
                      // if its intensity is above the threshold.

                      if(m_curApex.second > m_threshold)
                        // The current apex was above the threshold
                        {
                          cluster_sp->push_back(
                            std::make_shared<IndexIntensity>(m_curApex));

                          qDebug().noquote()
                            << "There was an apex already inside of cluster "
                               "range. "
                               "AND, since the current apex was above the "
                               "threshold, we indeed push it to cluster.\n";

                          qDebug().noquote() << "Current apex PUSHED: "
                                             << trace[m_curApex.first].x << ", "
                                             << m_curApex.second;

                          m_prevApex = m_curApex;

                          qDebug() << "We set prev apex to be cur apex.";
                        }
                      else
                        {
                          qDebug().noquote()
                            << "There was an apex already inside of cluster "
                               "range. "
                               "BUT, since the current apex was below the "
                               "threshold, we do not push it to cluster.\n";

                          qDebug().noquote() << "Current apex NOT pushed: "
                                             << trace[m_curApex.first].x << ", "
                                             << m_curApex.second;
                        }
                    }
                }
              else
                // No apex was previously found. We are fillin-up a new isotopic
                // cluster.
                {
                  if(m_curApex.second > m_threshold)
                    // We can actually add that apex to a new isotopic cluster.
                    {
                      if(cluster_sp->size() != 0)
                        qCritical(
                          "At this point, the cluster should be new and "
                          "empty.");

                      cluster_sp->push_back(
                        std::make_shared<IndexIntensity>(m_curApex));

                      qDebug().noquote()
                        << "No previous apex was found. Since current apex' "
                           "intensity is above threshold, we push it back to "
                           "the "
                           "cluster.\n";

                      // Store current apex as previous apex for next rounds.
                      m_prevApex = m_curApex;

                      qDebug().noquote() << "And thus we store the current "
                                            "apex as previous apex.\n";
                    }
                  else
                    {
                      qDebug().noquote()
                        << "No previous apex was found. Since current apex' "
                           "intensity is below threshold, we do nothing.\n";
                    }
                }
            }
          // End of
          // ! if(!was_ascending_to_apex)
          // That is, we were ascending to an apex.

          // Tell what it is: we were not ascending to an apex.
          was_ascending_to_apex = false;
        }
      // End of
      // ! if(m_cur.second > m_prev.second)
      // That is, we are descending a peak.

      // At this point, prepare next round.

      qDebug().noquote()
        << "Preparing next round, with m_prev = m_cur and ++index.\n";

      m_prev = m_cur;
      ++index;
    }
  // End of
  // while(index < trace_size)

  // At this point, if a cluster had been cooking, add it.

  if(cluster_sp->size())
    m_isotopicClusters.push_back(cluster_sp);

  return m_isotopicClusters.size();
}


Trace &
FilterLowIntensitySignalRemoval::filter(Trace &trace) const
{
  qDebug();

  // Horrible hack to have a non const filtering process.
  return const_cast<FilterLowIntensitySignalRemoval *>(this)->nonConstFilter(
    trace);
}


Trace &
FilterLowIntensitySignalRemoval::nonConstFilter(Trace &trace)
{
  qDebug();

  // Make a copy of the trace.
  m_helperTraceCopy = Trace(trace);

  if(trace.size() <= 2)
    {
      qDebug() << "The original trace has less than 3 points. Returning it "
                  "without modification.";
      return trace;
    }
  else
    qDebug() << "The original trace had" << trace.size() << "data points";

  std::size_t isotopic_cluster_count = detectIsotopicClusters(trace);

  qDebug() << "Number of isotopic clusters: " << isotopic_cluster_count;

  Trace final_trace;

  for(auto &&apices_p : m_isotopicClusters)
    {
      qDebug() << "iterating in new cluster.";

      for(auto &&apex_p : *apices_p)
        {
          DataPoint data_point =
            DataPoint(trace[apex_p->first].x, apex_p->second);

          qDebug() << "Current cluster data point:" << data_point.toString();

          final_trace.push_back(data_point);
        }
    }

  trace = std::move(final_trace);

  return trace;
}


double
FilterLowIntensitySignalRemoval::getThreshold() const
{
  return m_threshold;
}


//! Return a string with the textual representation of the configuration data.
QString
FilterLowIntensitySignalRemoval::toString() const
{
  return QString("%1|%2").arg(name()).arg(QString::number(m_threshold, 'f', 2));
}


QString
FilterLowIntensitySignalRemoval::name() const
{
  return "FilterLowIntensitySignalRemoval";
}

} // namespace pappso
